package mandelbrot;

import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;

/**
 * Parallel implementation of the Mandelbrot set computation (Concrete class in
 * the Template Method pattern).
 * 
 * @author ahoefer
 * 
 */
public class ParallelMandelbrot extends AbstractMandelbrot {

	private int numberOfThreads;

	/**
	 * The created algorithm will compute a Mandelbrot set with the given width
	 * and height.
	 * 
	 * @param width
	 *            the width of the Mandelbrot
	 * @param height
	 *            the height of the Mandelbrot
	 * 
	 * @param numberOfThreads
	 *            the number of threads used for computation
	 */
	public ParallelMandelbrot(int width, int height, int numberOfThreads) {
		super(width, height);
		this.numberOfThreads = numberOfThreads;
	}

	/**
	 * Computes the Mandelbrot set in parallel.
	 * 
	 * @return the Mandelbrot
	 */
	@Override
	protected int[][] compute() {
		int w = (int) Math.ceil((double) width / numberOfThreads);

		CyclicBarrier barrier = new CyclicBarrier(numberOfThreads + 1);
		Runnable r;
		Thread[] threads = new Thread[numberOfThreads];
		int[][] result = new int[width][height];
		for (int i = 0; i < numberOfThreads; i++) {
			int fromW = i * w;
			int toW = Math.min((i + 1) * w, width);
			r = new RunnableMandelbrot(barrier, fromW, toW, 0, height, result);
			threads[i] = new Thread(r);
			threads[i].start();
		}

		try {
			barrier.await();
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (BrokenBarrierException e) {
			e.printStackTrace();
		}

		return result;
	}

	private class RunnableMandelbrot implements Runnable {
		private int fromH;
		private int toH;
		private int fromW;
		private int toW;
		private int[][] target;
		private CyclicBarrier barrier;

		public RunnableMandelbrot(CyclicBarrier barrier, int fromW, int toW,
				int fromH, int toH, int[][] target) {
			this.barrier = barrier;
			this.fromH = fromH;
			this.toH = toH;
			this.fromW = fromW;
			this.toW = toW;
			this.target = target;
		}

		public void run() {
			double cReal;
			double cImag;
			for (int x = fromW; x < toW; x++) {
				for (int y = fromH; y < toH; y++) {
					cReal = translateReal(x);
					cImag = translateImag(y);
					target[x][y] = computeMandelbrotPoint(cReal, cImag);
				}
			}
			try {
				barrier.await();
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (BrokenBarrierException e) {
				e.printStackTrace();
			}
		}
	}
}
